//////////////////////////////////////////////////////////////////
//
// ICQMsg.c - ICQ Message processing stuff.
// 
//   Copyright (C) 1998 Paul Holden
//   paulh@javigate.com
// 
//   File Created: 6  Nov 1998
//   Last Update: 16 Nov 1998
//
//   For more info and more recent version see:
//     http://www.javigate.com/ICQ/
//
//  Standard Disclaimer:
//  --------------------
//    Note that I am in no way affiliate with Mirabilis. I have found all these
//    information by tracing UDP packets (this complies with Mirabilis License
//    agreement). This information are unofficial and may be incorrect.
//
//    This program is provided "as is" and comes with no warranty,
//    express if implied. Use this program at you own risk. Yadda Yadda.


#define WIN32_LEAN_AND_MEAN		// Dont' include a load of crap
#include <windows.h>			// We need windows for defines of DWORD etc
#include <winsock2.h>			// And Winsock (for inet_ntoa etc)
#include <stdlib.h>				// free/malloc
#include <stdio.h>				// sprintf

#include "ISpyUMain.h"
#include "ICQMsg.h"
#include "SocketHandler.h"

// Allocate storage for a new message.
// If this can't allocate enough memory for something,
// it frees everything it has allocated and returns NULL

// We use malloc but in the future I'll either:
// a) Convert it to C++ and use "new"
// or
// b) Use LocalAlloc
ICQMsg *
CreateICQMsg(unsigned long nMsgLen, BOOL bServerMsg) {
	ICQMsg *pICQMsg;

	pICQMsg = (ICQMsg *)malloc(sizeof(ICQMsg));
	if (pICQMsg == NULL) 
		return NULL;

	pICQMsg->pData			= NULL;
	pICQMsg->pDecryptedData = NULL;
	pICQMsg->nMsgLen		= nMsgLen;
	pICQMsg->bServerMsg		= bServerMsg;
	pICQMsg->bDecrypted		= FALSE;
	pICQMsg->pNextMsg		= NULL;
	pICQMsg->pPrevMsg		= NULL;

	// Allocate buffers for data
	pICQMsg->pData          = (unsigned char *)malloc(nMsgLen+40);	// Give us some extra breathing space
	pICQMsg->pDecryptedData = (unsigned char *)malloc(nMsgLen+40);	// (the decryption routine below splatters
																	// the last few byte of the buffer)

	// If either of these allocs failed, free whatever we have allocated
	// and return NULL
	if (pICQMsg->pData == NULL || pICQMsg->pDecryptedData == NULL) {
		DestroyICQMsg(pICQMsg);
		return NULL;
	}
	return pICQMsg;
}

// Deletes the message and associated memory.
// Doesn't check if this message is in a list or not..
void
DestroyICQMsg(ICQMsg *pICQMsg) {
	if (pICQMsg->pDecryptedData)
		free(pICQMsg->pDecryptedData);
	if (pICQMsg->pData)
		free(pICQMsg->pData);

	free(pICQMsg);
}

void
DestroyICQMsgList(ICQMsg **pICQMsgHead) {
	ICQMsg *pICQMsg = *pICQMsgHead;
	ICQMsg *pNextMsg;

	while (pICQMsg) {
		pNextMsg = pICQMsg->pNextMsg;
		RemoveICQMsg(pICQMsgHead, pICQMsg);
		DestroyICQMsg(pICQMsg);
		pICQMsg = pNextMsg;
	}
}

// Add a message to the head of the linked list.
// I use doubly-linked lists, as I prefer them
// over single-linked lists. No other reason...
void
AddICQMsgTail(ICQMsg **pICQMsgHead, ICQMsg *pICQMsg) {
	ICQMsg *pLastMsg;

	pLastMsg = *pICQMsgHead;
	if (pLastMsg == NULL) {
		// No other messages...
		pICQMsg->pPrevMsg = NULL;
		pICQMsg->pNextMsg = NULL;
		*pICQMsgHead = pICQMsg;
	} else {
		// Get pointer to last message
		while(pLastMsg->pNextMsg != NULL)
			pLastMsg = pLastMsg->pNextMsg;

		// pLastMsg now points to last message in list...
		pICQMsg->pNextMsg = NULL;
		pICQMsg->pPrevMsg = pLastMsg;
		pLastMsg->pNextMsg = pICQMsg;
	}
}

// Add a message to the head of the linked list.
void
AddICQMsgHead(ICQMsg **pICQMsgHead, ICQMsg *pICQMsg) {
	
	pICQMsg->pNextMsg = *pICQMsgHead;
	pICQMsg->pPrevMsg = NULL;
	if (*pICQMsgHead) {
		(*pICQMsgHead)->pPrevMsg = pICQMsg;
	}
	*pICQMsgHead = pICQMsg;
}

// Removes a message from the linked list of messages.
void
RemoveICQMsg (ICQMsg **pICQMsgHead, ICQMsg *pICQMsg) {

	ICQMsg *pMsg, *pNextMsg, *pPrevMsg;
	
	pMsg = *pICQMsgHead;
	while (pMsg) {
		if (pMsg == pICQMsg) {
			// We've found a match
			pNextMsg = pMsg->pNextMsg;
			pPrevMsg = pMsg->pPrevMsg;
			
			if (pPrevMsg) {
				pPrevMsg->pNextMsg = pNextMsg;
			} else {
				// This is first in list
				*pICQMsgHead = pNextMsg;
			}

			if (pNextMsg) {
				pNextMsg->pPrevMsg = pPrevMsg;
			}
			return;
		} 
		pMsg = pMsg->pNextMsg;
	}
}

// Print the time that this message was received to 
// a buffer. Change this if you don't like the date
// format :-)
void
GetICQMsgTime(ICQMsg *pICQMsg, char *lpBuffer) {
	time_t lTime = pICQMsg->lTime;
	struct tm *today = localtime(&lTime);

    strftime( lpBuffer, 128, "%d %b %H:%M:%S", today );
}

// This prints information about a message to the specified string.
// At the moment the information is too verbose (so you can use
// GetICQMsgHeaderInfo below)
void
GetICQMsgInfo(ICQMsg *pICQMsg, char *szBuffer) {
	char szCommand[256] = "";

	GetCommandName(ICQCommand(pICQMsg), szCommand);

	if (!pICQMsg->bServerMsg) {
		sprintf(szBuffer, "ClientMsg:\r\n" \
						  "  Version\t%04lx\r\n" \
						  "  Zero\t\t%08lx\r\n" \
						  "  nUIN\t\t%08lx\t(%d)\r\n" \
						  "  Session\t%08lx\r\n" \
						  "  Command\t%04lx\t\t(%s)\r\n" \
						  "  SeqNum1/2\t%04lx/%04lx\r\n" \
						  "  nCheckCode\t%08lx\r\n",
						  ICQVersion(pICQMsg),
						  ICQZero(pICQMsg),
						  ICQUIN(pICQMsg),ICQUIN(pICQMsg),
						  ICQSession(pICQMsg),
						  ICQCommand(pICQMsg), szCommand,
						  ICQSeqNum1(pICQMsg),
						  ICQSeqNum2(pICQMsg),
						  ICQCheckCode(pICQMsg));
	} else {
		sprintf(szBuffer, "ServerMsg:\r\n" \
						  "  Version\t%04lx\r\n" \
						  "  Zero\t\t%08lx\r\n" \
						  "  nUIN\t\t%08lx\t(%d)\r\n" \
						  "  Session\t%08lx\r\n" \
						  "  Command\t%04lx\t\t(%s)\r\n" \
						  "  SeqNum1/2\t%04lx/%04lx\r\n" \
						  "  nCheckCode\t%08lx\r\n",
						  ICQVersion(pICQMsg),
						  ICQZero(pICQMsg),
						  ICQUIN(pICQMsg),ICQUIN(pICQMsg),
						  ICQSession(pICQMsg),
						  ICQCommand(pICQMsg), szCommand,
						  ICQSeqNum1(pICQMsg), ICQSeqNum2(pICQMsg),
						  ICQCheckCode(pICQMsg));
	}
	strcat(szBuffer, pICQMsg->szExtraInfo);
	strcat(szBuffer, "\r\n");
}

// Find the (first) message with the given Session and SeqNums
// If SeqNum1/2 is -1, then we ignore them (a bit like specifying a wildcard...)
ICQMsg *
FindICQMsg(ICQMsg **pICQMsgHead, DWORD nSession, WORD nSeqNum1, WORD nSeqNum2) {
	ICQMsg *pMsg;

	pMsg = *pICQMsgHead;
	while(pMsg) {
		if (ICQCommand(pMsg) != ICQ_ACK) {
			if ((ICQSeqNum1(pMsg) == nSeqNum1 || (WORD)nSeqNum1 == (WORD)-1) &&
				(ICQSeqNum2(pMsg) == nSeqNum2 || (WORD)nSeqNum2 == (WORD)-1) &&
				ICQSession(pMsg) == nSession)

				return pMsg;
		}
		pMsg = pMsg->pNextMsg;
	}
	return NULL;

}

// Print the header of the given Server/Client message to a string.
// There's a bit of redundancy here, and it displays the byte
// ordering backwards. I'll fix this pretty soon I think.
// This routine was useful when I was trying to figure out how the
// check code was calculated for the Server messages, but it may
// be rather useless now.
void
GetICQMsgHeaderInfo(ICQMsg **pICQMsgHead, ICQMsg *pICQMsg, char *szBuffer) {
	char szCommandName[256];
	ICQMsg *pMsg;

	if (pICQMsg->bServerMsg) {
		GetCommandName(ICQCommand(pICQMsg), szCommandName);

		// Display first however many bytes....
		//				         Ver  0     Session   Comm   Seq1/2      UIN   Check
		sprintf(szBuffer, "S: v:%d z:%02d s:%08lx c:%s sq:%d/%d #:%d x:%08lx ",
			ICQVersion(pICQMsg),
			ICQZero(pICQMsg),
			ICQSession(pICQMsg),
			szCommandName,
			ICQSeqNum1(pICQMsg),
			ICQSeqNum2(pICQMsg),
			ICQUIN(pICQMsg),
			ICQCheckCode(pICQMsg));

		// If this message is an ACK (ie Acknowledgement), we try to show which
		// message the ACK was in responce to. Not really all that useful now
		if (ICQCommand(pICQMsg) == ICQ_ACK) {

			strcat(szBuffer, "ACK: ");
			if (pMsg = FindICQMsg(pICQMsgHead, ICQSession(pICQMsg), ICQSeqNum1(pICQMsg), ICQSeqNum2(pICQMsg)))
				GetCommandName(ICQCommand(pMsg), szCommandName);
			else 
				strcpy(szCommandName, "(Unknown)");

			// Concatenate with this buffer
			strcat(szBuffer, szCommandName);
		} 

	} else {
		GetCommandName(ICQCommand(pICQMsg), szCommandName);

		// Display first however many bytes....
		//				       Ver  0      UID   SESS   COMM  SEQ1/2  CHECK
		sprintf(szBuffer, "C: v:%d z:%08lx #:%d s:%08lx c:%s sq:%d/%d x:%08lx ",
			ICQVersion(pICQMsg),
			ICQZero(pICQMsg),
			ICQUIN(pICQMsg),
			ICQSession(pICQMsg),
			szCommandName,

			ICQSeqNum1(pICQMsg),
			ICQSeqNum2(pICQMsg),
			ICQCheckCode(pICQMsg));

		// If this message is an ACK (ie Acknowledgement), we try to show which
		// message the ACK was in responce to. Not really all that useful now
		if (ICQCommand(pICQMsg) == ICQ_ACK) {

			strcat(szBuffer, "ACK: ");
			// For some reason, when cleint acknowledges server, SeqNum2 is always 0.?!?!
			// We set SeqNum2 to -1 to allow us to ignore the SeqNum2 of server messages when searching for it
			if (pMsg = FindICQMsg(pICQMsgHead, ICQSession(pICQMsg), ICQSeqNum1(pICQMsg), -1))
				GetCommandName(ICQCommand(pMsg), szCommandName);
			else 
				strcpy(szCommandName, "(Unknown)");

			// Concatenate with this buffer
			strcat(szBuffer, szCommandName);
		} 
	}
}

// Obtain the key used by the client to "encrypt" its data
// This involves descrambling the bits in the CheckCode, 
// and adding it to the product of the
// packet length and a magic number
DWORD
GetKey(ICQMsg *pICQMsg) {
	DWORD key = 0, check;
		DWORD A1,A2,A3,A4,A5;

	if (ICQVersion(pICQMsg) == 5) {
		// Unscramble bits

		check = ICQCheckCode(pICQMsg);

		A1 = check & 0x0001F000;
		A2 = check & 0x07C007C0;
		A3 = check & 0x003E0001;
		A4 = check & 0xF8000000;
		A5 = check & 0x0000083E;

		A1 = A1 >> 0x0C;
		A2 = A2 >> 0x01;
		A3 = A3 << 0x0A;
		A4 = A4 >> 0x10;
		A5 = A5 << 0x0F;

		check = A1 + A2 + A3 + A4 + A5;

		key = 0x68656C6C * pICQMsg->nMsgLen;
		key += check;

	} else if (ICQVersion(pICQMsg) == 4) {
		check = ICQCheckCode(pICQMsg) ^ 0x00ff00ff;

		/*A1 = check & 0x0001F000;
		A2 = check & 0x07C007C0;
		A3 = check & 0x003E0001;
		A4 = check & 0xF8000000;
		A5 = check & 0x0000083E;

		A1 = A1 >> 0x0C;
		A2 = A2 >> 0x01;
		A3 = A3 << 0x0A;
		A4 = A4 >> 0x10;
		A5 = A5 << 0x0F;

		check = A1 + A2 + A3 + A4 + A5;*/

		key = 0x66756B65 * pICQMsg->nMsgLen;
		key += check;
	} 
	return key;
}

// Take a given DWORD, and reverse the byte ordering.
// For Motorola processors, this doesn't need to be
// Done. Or something (I get confused by all big/little endian
// things - probably because I spent 3 year writing 68000
// assembly languange and everything to do with Intel assembly
// confuses me :-)
DWORD
Reverse(DWORD dw)
{
	unsigned char z[4];

	z[0] = (unsigned char)((dw)>>24)& 0x000000FF;
	z[1] = (unsigned char)((dw)>>16)& 0x000000FF;
	z[2] = (unsigned char)((dw)>>8) & 0x000000FF;
	z[3] = (unsigned char)(dw)      & 0x000000FF;

	return *((DWORD *)z);
}

// This routine takes the message data in pICQMsg
// and decrypts it, storing the decrypted data
// in pDecryptedData.
//  It this message is from the server, this routine
// simply copies the data from PData to pDecryptedData
// as the server doesn't "encrypt" data (something
// else I didn't realsie for a day or so - d'oh! :-)
void
DecryptPacket(ICQMsg *pICQMsg) {
	DWORD nCode = 0, nData, key;

	int n, pos, t;

	// Skip if we've already decrpyted this message.
	// I was going to call this function whenever
	// decrypted data needed to be read, so this stopped
	// from decrypting the message more than once.
	// Not reall needed now, but it may come in useful later/
	if (pICQMsg->bDecrypted == TRUE) return;
	pICQMsg->bDecrypted = TRUE;

	// Copy pData to pDecryptedData buffer.
	// If this is a Server message, we can leave after this bit
	memcpy(pICQMsg->pDecryptedData, pICQMsg->pData, pICQMsg->nMsgLen);

	// Don't decrypt Server msgs or messages that aren't v4 or v5
	if (pICQMsg->bServerMsg ||
		(ICQVersion(pICQMsg) != 4 &&
		 ICQVersion(pICQMsg) != 5) ) return;

	// We need to get the key for this message. 
	// The key is calculated above (it involves unscrambing
	// the CheckCode's bits, and adding that value to
	// the packet length multiplied by a magic number.
	key = GetKey(pICQMsg);


	if (ICQVersion(pICQMsg) == 4) {
		int count, i;
		count =(pICQMsg->nMsgLen + 3) / 4;
		count+=3;
		count/=4;
		count = 3;
 
		for(i=0; i<count; i++) {
			t = (i*4) % 256;
			nCode = key + icq_check_data5[t*4];

			if(i!=4) {	// checksum is not encrypted
				nData = *(DWORD *)(&pICQMsg->pDecryptedData[i*4]);

				nData ^=  nCode;

				*(DWORD *)(&pICQMsg->pDecryptedData[i*4]) = nData;
			}
		}
		// Replace first two bytes...
		*(WORD *)(pICQMsg->pDecryptedData) = 4;

	}
	// This routine doesn't actually work right - 
	// it splatters the last few bytes of the buffer.
	// Will be fixed soon.
	// Basically loops through each DWORD in the buffer,
	// XORing each with nCode (which is the key value
	// plus a value in the magic check table.

	if (ICQVersion(pICQMsg) == 5) {
		n = (pICQMsg->nMsgLen + 3);

		pos = 0x0A;					// Start at position 10

		while (pos < n) {
			t = pos % 256;		// There are 256 values in the table...

			nCode = key + icq_check_data5[t];

			nData = *(DWORD *)(&pICQMsg->pDecryptedData[pos]);

			nData ^= nCode;				// XOR with code

			*(DWORD *)(&pICQMsg->pDecryptedData[pos]) = nData;

			pos += 4;
		}
	}


}

// This creates a valid check code as used by the ICQ server
// to stamp messages with. It's pretty similar to the routine
// below to generate ClientCheck codes, but it took me about
// 2 days and about 100 ACK packets to figure out.
//  Basically, the only difference is that the server doesn't
// scramble the bits at the end to obfuscate things.
//  It operates on pData, and not pDecryptedData, as the server
// never actually encrypts its messages.
// 
DWORD
CreateServerCode(ICQMsg *pICQMsg) {

	BYTE B2, B4, B6, B8;
	DWORD NUMBER1, NUMBER2;
	BYTE X1, X2, X3, X4;

	B2 = *((BYTE  *)(pICQMsg->pData+2));
	B4 = *((BYTE  *)(pICQMsg->pData+4));
	B6 = *((BYTE  *)(pICQMsg->pData+6));
	B8 = *((BYTE  *)(pICQMsg->pData+8));

	NUMBER1 = (B8<<24) | (B4<<16) | (B2<<8) | B6;

	X4 = 5;
	X3 = ~ *((BYTE  *)(pICQMsg->pData+X4));
	X2 = rand() % 255;
	X1 = ~ icq_check_data5[X2];

	NUMBER2 = (X4<<24) | (X3<<16) | (X2<<8) | X1;

	return NUMBER1 ^ NUMBER2;
}

// This creates a check code as described by
// Sebastien Dault (daus01@gel.usherb.ca)
DWORD
CreateCheckCode(ICQMsg *pICQMsg, BYTE R1, BYTE R2) {

	BYTE B2, B4, B6, B8;
	DWORD NUMBER1, NUMBER2;
	BYTE X1, X2, X3, X4;
	DWORD A1, A2, A3, A4, A5;
	DWORD CHECK;

	B2 = *((BYTE  *)(pICQMsg->pDecryptedData+2));
	B4 = *((BYTE  *)(pICQMsg->pDecryptedData+4));
	B6 = *((BYTE  *)(pICQMsg->pDecryptedData+6));
	B8 = *((BYTE  *)(pICQMsg->pDecryptedData+8));

	NUMBER1 = (B8<<24) | (B4<<16) | (B2<<8) | B6;
	NUMBER2 = NUMBER1 ^ *((DWORD *)(pICQMsg->pDecryptedData+11));

	// Uncommenting these lines will cause this function to
	// generate a random checkcode each time
	//R1 = (BYTE)(rand() % (pICQMsg->nMsgLen - 18));
	//R2 = (BYTE)(rand() % 0x100);

	X4 = R1;
	X3 = ~ *((BYTE  *)(pICQMsg->pDecryptedData+X4));

	X2 = R2;
	X1 = ~ icq_check_data5[X2];

	// X1 should equal hi (or low?) byte of NUMBER2

	NUMBER2 = (X4<<24) | (X3<<16) | (X2<<8) | X1;

	CHECK = NUMBER1 ^ NUMBER2;

	A1 = CHECK & 0x0000001F;
	A2 = CHECK & 0x03E003E0;
	A3 = CHECK & 0xF8000400;
	A4 = CHECK & 0x0000F800;
	A5 = CHECK & 0x041F0000;

	A1 = A1 << 0x0C;
	A2 = A2 << 0x01;
	A3 = A3 >> 0x0A;
	A4 = A4 << 0x10;
	A5 = A5 >> 0x0F;

   return A1 + A2 + A3 + A4 + A5;
}

// This is just a routine I knocked up to test
// every possible value of R1/R2 and find out
// what random numbers the Client used to
// "encrypt" the message with (to see if there
// was a pattern).
// Uncomment the MessageBox lines to make it
// display what it discovers.
BOOL
TestCheckCode(ICQMsg *pICQMsg) {
	WORD R1, R2;
	DWORD CHECK;
	char szBuffer[256];

	DWORD check;
	
	return TRUE;
	
	check = ICQCheckCode(pICQMsg);

	// Create a check code for every possible value of R1 and R2..
	for (R1 = 0; R1 < pICQMsg->nMsgLen - 18; R1++) {
		for (R2 = 0; R2 < 256; R2++) {
			CHECK = CreateCheckCode(pICQMsg, (BYTE)R1, (BYTE)R2);

			if (CHECK == check) {
				sprintf(szBuffer, "R1 = %04x, R2 = %04x", R1, R2);
				//MessageBox(NULL, szBuffer, "Found Random Values", MB_OK);
				return TRUE;
			}
		}
	}
	//MessageBox(NULL, "Failed to find magic values", "Arse", MB_OK);
	return FALSE;
}

/*
// This routine initialises the header information
// on a v5 packet (no v4 as yet...)
// Also, as we don't currently send spoof client messages,
// this only initialises ICQSERVER_MESSAGE headers...

05 00			VERSION
ZZ				ZERO (possibly used to affect CheckCode...?)
II II II II		SESSION# (usually copied from session of client message)
CC CC			COMMAND (the command we want to use)
SS SS			SEQNUM1
TT TT			SEQNUM2
UU UU UU UU		UIN
XX XX XX XX		CHECK
*/

void
InitHeader(ICQMsg *pICQMsg,
		   WORD nVersion,
		   DWORD nZero,
		   DWORD nSession,
		   WORD  nCommand,
		   WORD  nSeqNum1,
		   WORD  nSeqNum2,
		   DWORD nUIN) {

	if (nVersion != 5) return;	// Don't do <v5 packets yet.

	if (pICQMsg->bServerMsg) {
		*((WORD  *)(pICQMsg->pData+0)) = nVersion;
		*((BYTE  *)(pICQMsg->pData+2)) = (BYTE)nZero;	// Zero is a BYTE for server packets, but a DWORD for client packets...
		*((DWORD *)(pICQMsg->pData+3)) = nSession;
		*((WORD  *)(pICQMsg->pData+7)) = nCommand;
		*((WORD  *)(pICQMsg->pData+9)) = nSeqNum1;
		*((WORD  *)(pICQMsg->pData+11))= nSeqNum2;
		*((DWORD *)(pICQMsg->pData+13))= nUIN;
		*((DWORD *)(pICQMsg->pData+17))= CreateServerCode(pICQMsg);
	}
}

// This creates a LOGIN_REPLY message
// Most of the fields are as yet unknown, so I send values ripped from
// packets I snooped with this program
ICQMsg *
CreateLoginReply(ICQMsg *pICQMsg) {
	ICQMsg *pNewICQMsg;

	// The LOGIN_REPLY messages are 53 bytes long..
	// ToDo: check ICQVersion of pICQMsg and create different LOGIN_REPLYs depending on
	// version. This assumes v5, but may work with other protcols.
	if (pNewICQMsg = CreateICQMsg(53, TRUE)) {

		// Fill in fields...
		InitHeader(pNewICQMsg, 0x005,					// v5
							   0,
							   ICQSession(pICQMsg),		// Use same session as client's message
							   ICQ_LOGIN_REPLY,			// The command type
							   0x0000,					// Our seqnums. At the moment, these
							   0x0001,					// are hardcoded. Should really change as program runs. ?
							   ICQUIN(pICQMsg));		// Copy Client's UIN

		*((DWORD *)(pNewICQMsg->pData+21))= Reverse(0xc2de0fdb);	// IP address - currently hard coded to MY IP (don't flood me! :-)
		*((DWORD *)(pNewICQMsg->pData+25))= Reverse(0x01000b00);	// Other values are copied directly from packet's I've intercepted
		*((DWORD *)(pNewICQMsg->pData+29))= Reverse(0x19001600);	// No idea what most of them are...
		*((DWORD *)(pNewICQMsg->pData+33))= Reverse(0x8c000000);
		*((DWORD *)(pNewICQMsg->pData+37))= Reverse(0xe0010a00);
		*((DWORD *)(pNewICQMsg->pData+41))= Reverse(0x0a000500);
		*((DWORD *)(pNewICQMsg->pData+45))= Reverse(0x0a000197);
		*((DWORD *)(pNewICQMsg->pData+49))= Reverse(0x0fac1201);                                  
	}
	// This will return NULL if CreateICQMsg failed...
	return pNewICQMsg;
}

// This creates a REPLY_X1 message. I currently don't know
// Anything about this message. Read notes for CreateLoginReply above.
ICQMsg *
CreateReply_X1(ICQMsg *pICQMsg) {
	ICQMsg *pNewICQMsg;

	// REPLY_X1 is 25 bytes in length
	if (pNewICQMsg = CreateICQMsg(25, TRUE)) {

		// Fill in fields...
		InitHeader(pNewICQMsg, 0x005,					// Currently only v5
							   0,
							   ICQSession(pICQMsg),
							   ICQ_REPLY_X1,
							   0x0002,					// Again sessions are hard coded - should be
							   0x0000,					// Assigned dynamically
							   ICQUIN(pICQMsg));

		*((DWORD *)(pNewICQMsg->pData+21))= ICQUIN(pICQMsg);	// 21 appears to be the user's UIN (confirmation??)

	}
	return pNewICQMsg;
}

// Create an ACK message.
ICQMsg *
CreateAck(ICQMsg *pICQMsg) {
	ICQMsg *pNewICQMsg;

	// If pICQMsg is a client message, this new message
	// is a server message, and vice versa
	BOOL bServerMsg = !pICQMsg->bServerMsg;

	// ACK is 21 bytes
	if (pNewICQMsg = CreateICQMsg(21, bServerMsg)) {

		// Fill in fields...
		InitHeader(pNewICQMsg, 0x005,
							   0,
							   ICQSession(pICQMsg),
							   ICQ_ACK,
							   ICQSeqNum1(pICQMsg),	// Set SeqNums to those of Client Msg
							   ICQSeqNum2(pICQMsg),
							   ICQUIN(pICQMsg));


	}
	return pNewICQMsg;
}

// Respond to a client message. This routine Spoofs server responses,
// by creating false messages and sending them to the client's port.
//
// Currently does very little but reply to LOGIN message so we can
// make client think it's successfully logged on. 
// Does send ACK messages to all non-ACK messages

// We pass in the head of the linked list so we
// can add more than one message to the list
void
RespondToMessage(ICQMsg *pICQMsg, struct sockaddr_in *replyto, char *szBuffer) {
	ICQMsg *pNewICQMsg;

	// Don't do anything with ACK messages for now....
	// Should probably check Seqnums and remove messages from 
	// a "NotAcknowledged" buffer
	if (ICQCommand(pICQMsg) == ICQ_ACK) {
		strcpy(szBuffer, "Got ACK message");
		return;
	}

	// First, send an ACKnowledgement - this assumes we're a 
	if (pNewICQMsg = CreateAck(pICQMsg)) {
		ProcessICQMsg(pNewICQMsg, NULL);
		SendUsingSpoofServSocket(pNewICQMsg, replyto);
		DestroyICQMsg(pNewICQMsg);
		sprintf(szBuffer, "Sent message to client");
	} else {
		sprintf(szBuffer, "Couldn't send message to client");
	}

	// Now, process message as necessary...
	if (ICQCommand(pICQMsg) == ICQ_LOGIN) {
		// For a login message, we need to send bot LOGIN_REPLY and REPLY_X1
		pNewICQMsg = CreateLoginReply(pICQMsg);
		if (pNewICQMsg) {
			ProcessICQMsg(pNewICQMsg, NULL);
			SendUsingSpoofServSocket(pNewICQMsg, replyto);
			DestroyICQMsg(pNewICQMsg);
		}
		
		pNewICQMsg = CreateReply_X1(pICQMsg);
		if (pNewICQMsg) {
			ProcessICQMsg(pNewICQMsg, NULL);
			SendUsingSpoofServSocket(pNewICQMsg, replyto);
			DestroyICQMsg(pNewICQMsg);
		}
	}
}

// Perform some processing on a new ICQMsg.
// Basically, this sets up the timestamp and
// from info, as well as decrypting client messages
// and setting up szExtraInfo
BOOL
ProcessICQMsg(ICQMsg *pICQMsg, struct sockaddr_in *from) {
	WORD nCommand;

    time( &pICQMsg->lTime );

	// Process a few details about this message

	// Decrypt this message so that we can access info about
	// it (such as command name etc...)
	if (pICQMsg->bDecrypted == FALSE) {
		DecryptPacket(pICQMsg);
	}

	// If from is NULL, it's a message we've created ourselves,
	// to send to the client.
	if (from != NULL) {
		sprintf(pICQMsg->szFrom, "%s:%d", inet_ntoa(from->sin_addr), from->sin_port);
		TestCheckCode(pICQMsg);
	} else {
		sprintf(pICQMsg->szFrom, "Spoof");
	}

	// At the moment, we haven't extracted any extra information 
	// about this message.
	strcpy(pICQMsg->szExtraInfo, "");


	// Check what type of packet, and get each one to fill in
	// szExtraInfo
	nCommand = ICQCommand(pICQMsg);
	switch (nCommand) {
		case ICQ_LOGIN:
			ProcessLogin(pICQMsg);
			return TRUE;
		case ICQ_ACK:
			ProcessAck(pICQMsg);
			return TRUE;
		case ICQ_CONTACT_LIST:
			ProcessContactList(pICQMsg);
			return TRUE;
		case ICQ_INSTANT_MESSAGE:
			ProcessInstantMessage(pICQMsg);
			return TRUE;
		case ICQ_SEND_TEXT_CODE:
			ProcessTextCode(pICQMsg);
			return TRUE;
	}
	return FALSE;
}

// This just converts a 32-bit IP address to a string (it
// seemed easier than using inet_ntoa() as we did above
void
IPToStr(DWORD nIP, char *szBuffer) {
	sprintf(szBuffer, "%d.%d.%d.%d",
		nIP&0xFF,
		(nIP>>8)&0xFF,
		(nIP>>16)&0xFF,
		(nIP>>24)&0xFF);
}

/*
// Process an ICQ_LOGIN message. 
// This fills some information extracted from the
// LOGIN message into msg->szExtraInfo so we can
// quickly see the paramters.
VV VV				VERSION
ZZ ZZ ZZ ZZ    		ZERO
UU UU UU UU			UIN
II II II II			SESSION # (random)
CC CC				COMMAND
SS SS				SEQNUM1
TT TT				SEQNUM2
XX XX XX XX			CHECKCODE
-----------
5a a5 47 36			? Seems to change...
** ** 00 00			Port (Mirabilis seem to use a 32 bit number for this - not 16...)

PL PL				PASSWORDLEN (including terminating null)
P1 P2 P3 .. PN 00	PASSWORD (text, variable)
a0 00 00 00			?
** ** ** ** 		USER_IP
04					?
00 00 00 00			? STATUS ?
04 00 00 00			?
00 00 00 00			?
b8 00 a0 00			?
*/
void
ProcessLogin(ICQMsg *pICQMsg) {

	char szBuffer[16];			// Used to get text representation of IP address in

	// I'll convert all these nasty 24+x+y+.. bits into something a bit less untidy...
	DWORD nPort  = *(DWORD *)((pICQMsg->pDecryptedData)+24+4);
	WORD  nLength= *(WORD  *)((pICQMsg->pDecryptedData)+24+8);
	DWORD nIP    = *(DWORD *)((pICQMsg->pDecryptedData)+24+10+nLength+4);

	IPToStr(nIP, szBuffer);

	// Display password etc...
	sprintf(pICQMsg->szExtraInfo, "UID: %d, " \
					  "IP:Port: %s:%d, " \
					  "Password: %s (%d chars)",
					   
					  ICQUIN(pICQMsg),
					  szBuffer, nPort, 
					  (pICQMsg->pDecryptedData)+24+10, nLength);

}

// Get extra information about an ICQ_ACK message.
// This just lists the SeqNums of the message that
// the Client/Server wants to acknowledge
void
ProcessAck(ICQMsg *pICQMsg) {

	// Display seqnums
	sprintf(pICQMsg->szExtraInfo, "SeqNums: %d,%d",
					  ICQSeqNum1(pICQMsg),
					  ICQSeqNum2(pICQMsg));

}

/*
// The CONTACT_LIST message seems to have the following structure
VV VV				VERSION
ZZ ZZ ZZ ZZ    		ZERO
UU UU UU UU			UIN
II II II II			SESSION # (random)
CC CC				COMMAND
SS SS				SEQNUM1
TT TT				SEQNUM2
XX XX XX XX			CHECKCODE
-----------
NC					NUMCONTACTS
U1 U1 U1 U1			UIN #1
U2 U2 U2 U2			UIN #2
U3 U3 U3 U3         UIN #3
...
UN UN UN UN			UIN #NC
*/

void
ProcessContactList(ICQMsg *pICQMsg) {
	BYTE nNumContacts  = *(BYTE *)((pICQMsg->pDecryptedData)+24);
	DWORD nUIN;
	int i;
	char szBuffer[32];

	strcpy(pICQMsg->szExtraInfo, "");

	// Get UIN for each person in the list
	for (i = 0; i < nNumContacts; i++) {

		// This is a bit hairy- we shoould really check that we're not
		// trying to access past the end of the message buffer..
		nUIN = *(DWORD *)((pICQMsg->pDecryptedData)+24+1+i*4);

		// Print decimal representation of their UIN to buffer and concatenate...
		sprintf(szBuffer, "%d ", nUIN);
		strcat(pICQMsg->szExtraInfo, szBuffer);
	}
}

/*
// I call this INSTANT_MESSAGE - I'm not really sure what it is. 
// May be a technique of disguising IP address by sending through server?
05 00			VERSION
ZZ				ZERO
II II II II		SESSION#
CC CC			COMMAND
SS SS			SEQNUM1
TT TT			SEQNUM2
UU UU UU UU		UIN
XX XX XX XX		CHECK
-----------
US US US US		UIN of sender
01 00			? Status 
ML ML 			MSGLEN
M1 M2 .. MN 00  Message text(variable length) 
*/
void
ProcessInstantMessage(ICQMsg *pICQMsg) {
	WORD nMsgSize = *(WORD *)((pICQMsg->pDecryptedData)+27);
	WORD nUnknown = *(WORD *)((pICQMsg->pDecryptedData)+25);
	DWORD nUIN    = *(DWORD *)((pICQMsg->pDecryptedData)+21);

	sprintf(pICQMsg->szExtraInfo, "#%d: '%s' (Unknown = %d)",
				nUIN,
				(pICQMsg->pDecryptedData)+27+2,
				nUnknown);
}

/*
// Process SEND_TEXT_CODE messages
VV VV				VERSION
ZZ ZZ ZZ ZZ    		ZERO
UU UU UU UU			UIN
II II II II			SESSION # (random)
CC CC				COMMAND
SS SS				SEQNUM1
TT TT				SEQNUM2
XX XX XX XX			CHECKCODE
-----------
TL TL				TEXTCODELEN
T1 T2 .. TN 00		Message Text (eg B_USER_DISCONNECTED) (varibale length)
05 00				UNKNOWN - Version?

*/

void
ProcessTextCode(ICQMsg *pICQMsg) {
	WORD nMsgSize = *(WORD *)((pICQMsg->pDecryptedData)+24);
	WORD nUnknown = *(WORD *)((pICQMsg->pDecryptedData)+24+2+nMsgSize);

	// I print out the unknown at the end of the message. Let me know if you get anything but 5.
	// Just a thought - it's probably info specific to the specified message
	sprintf(pICQMsg->szExtraInfo, "'%s' (Unknown = %d)",
				(pICQMsg->pDecryptedData)+24+2,
				nUnknown);
}


// Check Code data used for v4 of the protocol
// Read this info from wumpus@innocent.com
// I got the v4 and v5 data from a couple of
// text files written by Sebastien Dault (daus01@gel.usherb.ca)

/*------------------------------------------------------------------------------

  ICQ "Secret" check data.

  This data is LIKELY to be copyrighted by ICQ.  This data is used with this
  program under the Fair Use clause of the U.S. Copyright Code.  

  The reason the use of this data falls under the Fair Use clause is that it
  is _NECESSARY_ for a program to use this data to interact with the ICQ 
  protocol.  Without this data, a program would not be able to successfully
  determine if a packet's "checksum" was valid, nor be able to communicate
  with the ICQ server.

  The reader might choose to draw their own conclusions about a company that
  needs to not only obscure a their protocol, but make it awkward for 3rd 
  parties to implement it.

 *----------------------------------------------------------------------------*/

/*
---   -----------------------------------------------    ----------------
 00 - 0A 5B 31 5D 20 59 6F 75 20 63 61 6E 20 6D 6F 64    .[1] You can mod
 10 - 69 66 79 20 74 68 65 20 73 6F 75 6E 64 73 20 49    ify the sounds I
 20 - 43 51 20 6D 61 6B 65 73 2E 20 4A 75 73 74 20 73    CQ makes. Just s
 30 - 65 6C 65 63 74 20 22 53 6F 75 6E 64 73 22 20 66    elect "Sounds" f
 40 - 72 6F 6D 20 74 68 65 20 22 70 72 65 66 65 72 65    rom the "prefere
 50 - 6E 63 65 73 2F 6D 69 73 63 22 20 69 6E 20 49 43    nces/misc" in IC
 60 - 51 20 6F 72 20 66 72 6F 6D 20 74 68 65 20 22 53    Q or from the "S
 70 - 6F 75 6E 64 73 22 20 69 6E 20 74 68 65 20 63 6F    ounds" in the co
 80 - 6E 74 72 6F 6C 20 70 61 6E 65 6C 2E 20 43 72 65    ntrol panel. Cre
 90 - 64 69 74 3A 20 45 72 61 6E 0A 5B 32 5D 20 43 61    dit: Eran.[2] Ca
 A0 - 6E 27 74 20 72 65 6D 65 6D 62 65 72 20 77 68 61    n't remember wha
 B0 - 74 20 77 61 73 20 73 61 69 64 3F 20 20 44 6F 75    t was said?  Dou
 C0 - 62 6C 65 2D 63 6C 69 63 6B 20 6F 6E 20 61 20 75    ble-click on a u
 D0 - 73 65 72 20 74 6F 20 67 65 74 20 61 20 64 69 61    ser to get a dia
 E0 - 6C 6F 67 20 6F 66 20 61 6C 6C 20 6D 65 73 73 61    log of all messa
 F0 - 67 65 73 20 73 65 6E 74 20 69 6E 63 6F 6D 69 6E    ges sent incomin
---   -----------------------------------------------    ----------------
*/


unsigned char icq_check_data4[256] = {
	0x0a, 0x5b, 0x31, 0x5d, 0x20, 0x59, 0x6f, 0x75, 0x20, 0x63, 0x61, 0x6e, 0x20, 0x6d, 0x6f, 0x64, 
	0x69, 0x66, 0x79, 0x20, 0x74, 0x68, 0x65, 0x20, 0x73, 0x6f, 0x75, 0x6e, 0x64, 0x73, 0x20, 0x49,
	0x43, 0x51, 0x20, 0x6d, 0x61, 0x6b, 0x65, 0x73, 0x2e, 0x20, 0x4a, 0x75, 0x73, 0x74, 0x20, 0x73, 
	0x65, 0x6c, 0x65, 0x63, 0x74, 0x20, 0x22, 0x53, 0x6f, 0x75, 0x6e, 0x64, 0x73, 0x22, 0x20, 0x66, 
	0x72, 0x6f, 0x6d, 0x20, 0x74, 0x68, 0x65, 0x20,	0x22, 0x70, 0x72, 0x65, 0x66, 0x65, 0x72, 0x65,
	0x6e, 0x63, 0x65, 0x73, 0x2f, 0x6d, 0x69, 0x73, 0x63, 0x22, 0x20, 0x69, 0x6e, 0x20, 0x49, 0x43,
	0x51, 0x20, 0x6f, 0x72, 0x20, 0x66, 0x72, 0x6f, 0x6d, 0x20, 0x74, 0x68, 0x65, 0x20, 0x22, 0x53,
	0x6f, 0x75, 0x6e, 0x64, 0x73, 0x22, 0x20, 0x69, 0x6e, 0x20, 0x74, 0x68, 0x65, 0x20, 0x63, 0x6f,
	0x6e, 0x74, 0x72, 0x6f, 0x6c, 0x20, 0x70, 0x61,	0x6e, 0x65, 0x6c, 0x2e, 0x20, 0x43, 0x72, 0x65,
	0x64, 0x69, 0x74, 0x3a, 0x20, 0x45, 0x72, 0x61,	0x6e, 0x0a, 0x5b, 0x32, 0x5d, 0x20, 0x43, 0x61,
	0x6e, 0x27, 0x74, 0x20, 0x72, 0x65, 0x6d, 0x65,	0x6d, 0x62, 0x65, 0x72, 0x20, 0x77, 0x68, 0x61,
	0x74, 0x20, 0x77, 0x61, 0x73, 0x20, 0x73, 0x61,	0x69, 0x64, 0x3f, 0x20, 0x20, 0x44, 0x6f, 0x75,
	0x62, 0x6c, 0x65, 0x2d, 0x63, 0x6c, 0x69, 0x63, 0x6b, 0x20, 0x6f, 0x6e, 0x20, 0x61, 0x20, 0x75,
	0x73, 0x65, 0x72, 0x20, 0x74, 0x6f, 0x20, 0x67,	0x65, 0x74, 0x20, 0x61, 0x20, 0x64, 0x69, 0x61,
	0x6c, 0x6f, 0x67, 0x20, 0x6f, 0x66, 0x20, 0x61, 0x6c, 0x6c, 0x20, 0x6d, 0x65, 0x73, 0x73, 0x61,
	0x67, 0x65, 0x73, 0x20, 0x73, 0x65, 0x6e, 0x74, 0x20, 0x69, 0x6e, 0x63, 0x6f, 0x6d, 0x69, 0x6e
};

unsigned char icq_check_data5[256] = {
	0x59, 0x60, 0x37, 0x6B, 0x65, 0x62, 0x46, 0x48, 0x53, 0x61, 0x4C, 0x59, 0x60, 0x57, 0x5B, 0x3D,
	0x5E, 0x34, 0x6D, 0x36, 0x50, 0x3F, 0x6F, 0x67, 0x53, 0x61, 0x4C, 0x59, 0x40, 0x47, 0x63, 0x39,
	0x50, 0x5F, 0x5F, 0x3F, 0x6F, 0x47, 0x43, 0x69, 0x48, 0x33, 0x31, 0x64, 0x35, 0x5A, 0x4A, 0x42,
	0x56, 0x40, 0x67, 0x53, 0x41, 0x07, 0x6C, 0x49, 0x58, 0x3B, 0x4D, 0x46, 0x68, 0x43, 0x69, 0x48,
	0x33, 0x31, 0x44, 0x65, 0x62, 0x46, 0x48, 0x53, 0x41, 0x07, 0x6C, 0x69, 0x48, 0x33, 0x51, 0x54,
	0x5D, 0x4E, 0x6C, 0x49, 0x38, 0x4B, 0x55, 0x4A, 0x62, 0x46, 0x48, 0x33, 0x51, 0x34, 0x6D, 0x36,
	0x50, 0x5F, 0x5F, 0x5F, 0x3F, 0x6F, 0x47, 0x63, 0x59, 0x40, 0x67, 0x33, 0x31, 0x64, 0x35, 0x5A,
	0x6A, 0x52, 0x6E, 0x3C, 0x51, 0x34, 0x6D, 0x36, 0x50, 0x5F, 0x5F, 0x3F, 0x4F, 0x37, 0x4B, 0x35,
	0x5A, 0x4A, 0x62, 0x66, 0x58, 0x3B, 0x4D, 0x66, 0x58, 0x5B, 0x5D, 0x4E, 0x6C, 0x49, 0x58, 0x3B,
	0x4D, 0x66, 0x58, 0x3B, 0x4D, 0x46, 0x48, 0x53, 0x61, 0x4C, 0x59, 0x40, 0x67, 0x33, 0x31, 0x64,
	0x55, 0x6A, 0x32, 0x3E, 0x44, 0x45, 0x52, 0x6E, 0x3C, 0x31, 0x64, 0x55, 0x6A, 0x52, 0x4E, 0x6C,
	0x69, 0x48, 0x53, 0x61, 0x4C, 0x39, 0x30, 0x6F, 0x47, 0x63, 0x59, 0x60, 0x57, 0x5B, 0x3D, 0x3E,
	0x64, 0x35, 0x3A, 0x3A, 0x5A, 0x6A, 0x52, 0x4E, 0x6C, 0x69, 0x48, 0x53, 0x61, 0x6C, 0x49, 0x58,
	0x3B, 0x4D, 0x46, 0x68, 0x63, 0x39, 0x50, 0x5F, 0x5F, 0x3F, 0x6F, 0x67, 0x53, 0x41, 0x25, 0x41,
	0x3C, 0x51, 0x54, 0x3D, 0x5E, 0x54, 0x5D, 0x4E, 0x4C, 0x39, 0x50, 0x5F, 0x5F, 0x5F, 0x3F, 0x6F,
	0x47, 0x43, 0x69, 0x48, 0x33, 0x51, 0x54, 0x5D, 0x6E, 0x3C, 0x31, 0x64, 0x35, 0x5A, 0x00, 0x00
};

// These are mini-functions to return the required parameter
// from the message (be it a Server or Client message).
// Some work need to be done to get them working correctly with v4 and below
WORD
ICQVersion(ICQMsg *pICQMsg)	{
	return *((WORD  *)((unsigned char *)pICQMsg->pData));
}



DWORD
ICQZero(ICQMsg *pICQMsg) {
	if (!pICQMsg->bServerMsg) 
		return *((DWORD *)((unsigned char *)pICQMsg->pDecryptedData+2));
	else
		return (DWORD)*((BYTE *)((unsigned char *)pICQMsg->pDecryptedData+2));
}

DWORD
ICQUIN(ICQMsg *pICQMsg) {
	if (!pICQMsg->bServerMsg) 
		if (ICQVersion(pICQMsg) == 4)
			return *((DWORD *)((unsigned char *)pICQMsg->pDecryptedData+12));
		else if (ICQVersion(pICQMsg) == 5)
			return *((DWORD *)((unsigned char *)pICQMsg->pDecryptedData+6));
		else
			return -1;
	else 
		return *((DWORD *)((unsigned char *)pICQMsg->pDecryptedData+13));
}

DWORD
ICQSession(ICQMsg *pICQMsg){
	if (!pICQMsg->bServerMsg) 
		if (ICQVersion(pICQMsg) == 5)
			return *((DWORD *)((unsigned char *)pICQMsg->pDecryptedData+10));
		else
			return 0;
	else
		return *((DWORD *)((unsigned char *)pICQMsg->pDecryptedData+3));
}

WORD
ICQCommand(ICQMsg *pICQMsg)	{
	if (!pICQMsg->bServerMsg) 
		if (ICQVersion(pICQMsg) == 4)
			return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+6));
		else if (ICQVersion(pICQMsg) == 5)
			return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+14));
		else
			return 0;
	else
		return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+7));
}

WORD
ICQSeqNum1(ICQMsg *pICQMsg)	{
	if (!pICQMsg->bServerMsg) 
		if (ICQVersion(pICQMsg) == 4)
			return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+8));
		else if (ICQVersion(pICQMsg) == 5)
			return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+16));
		else
			return 0;

	else
		return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+9));
}

WORD
ICQSeqNum2(ICQMsg *pICQMsg)	{
	if (!pICQMsg->bServerMsg) 
		if (ICQVersion(pICQMsg) == 4)
			return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+10));
		else if (ICQVersion(pICQMsg) == 5)
			return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+18));
		else
			return 0;
	else
		return *((WORD  *)((unsigned char *)pICQMsg->pDecryptedData+11));
}

DWORD
ICQCheckCode(ICQMsg *pICQMsg) {
	if (!pICQMsg->bServerMsg) 
		if (ICQVersion(pICQMsg) == 4)
			return *((DWORD  *)((unsigned char *)pICQMsg->pData+20));
		else if (ICQVersion(pICQMsg) == 5)
			return *((DWORD  *)((unsigned char *)pICQMsg->pData+20));
		else
			return 0;
	else
		return *((DWORD *)((unsigned char *)pICQMsg->pData+17));
}

// A kludgy way of getting the text representation of the command.
// I'll put this in a lookup table soon (honest 'gov)
void
GetCommandName(WORD nCommand, char *szBuffer) {

	switch (nCommand) {

		case ICQ_ACK:
			strcpy(szBuffer, "ACK");
			break;
		case ICQ_SEND_MESSAGE:
			strcpy(szBuffer, "SEND_MESSAGE");
			break;
		case ICQ_LOGIN:
			strcpy(szBuffer, "LOGIN");
			break;
		case ICQ_CONTACT_LIST:
			strcpy(szBuffer, "CONTACT_LIST");
			break;
		case ICQ_SEARCH_UIN:
			strcpy(szBuffer, "SEARCH_UIN");
			break;
		case ICQ_SEARCH_USER:
			strcpy(szBuffer, "SEARCH_USER");
			break;
		case ICQ_KEEP_ALIVE:
			strcpy(szBuffer, "KEEP_ALIVE");
			break;
		case ICQ_SEND_TEXT_CODE:
			strcpy(szBuffer, "SEND_TEXT_CODE");
			break;
		case ICQ_LOGIN_1:
			strcpy(szBuffer, "LOGIN_1");
			break;
		case ICQ_INFO_REQ:
			strcpy(szBuffer, "INFO_REQ");
			break;
		case ICQ_EXT_INFO_REQ:
			strcpy(szBuffer, "EXT_INFO_REQ");
			break;
		case ICQ_CHANGE_PASSWORD:
			strcpy(szBuffer, "CHANGE_PASSWORD");
			break;
		case ICQ_STATUS_CHANGE:
			strcpy(szBuffer, "STATUS_CHANGE");
			break;
		case ICQ_LOGIN_2:
			strcpy(szBuffer, "LOGIN_2");
			break;

		case ICQ_UPDATE_INFO:
			strcpy(szBuffer, "UPDATE_INFO");
			break;
		case ICQ_UPDATE_EXT_INFO:
			strcpy(szBuffer, "UPDATE_EXT_INFO");
			break;
		case ICQ_ADD_TO_LIST:
			strcpy(szBuffer, "ADD_TO_LIST");
			break;
		case ICQ_REQ_ADD_TO_LIST:
			strcpy(szBuffer, "REQ_ADD_TO_LIST");
			break;
		case ICQ_QUERY_SERVERS:
			strcpy(szBuffer, "QUERY_SERVERS");
			break;
		case ICQ_QUERY_ADDONS:
			strcpy(szBuffer, "QUERY_ADDONS");
			break;
		case ICQ_NEW_USER_1	:
			strcpy(szBuffer, "NEW_USER_1");
			break;
		case ICQ_NEW_USER_REG:
			strcpy(szBuffer, "NEW_USER_REG");
			break;
		case ICQ_NEW_USER_INFO:
			strcpy(szBuffer, "NEW_USER_INFO");
			break;
		case ICQ_CMD_X1:
			strcpy(szBuffer, "CMD_X1");
			break;
		//case ICQ_MSG_TO_NEW_USER:					// Hey! this message number is used
		//	strcpy(szBuffer, "MSG_TO_NEW_USER");	// Twice. Has anyone else realised this?
		//	break;


		case ICQ_LOGIN_REPLY:
			strcpy(szBuffer, "LOGIN_REPLY");
			break;
		case ICQ_USER_ONLINE:
			strcpy(szBuffer, "USER_ONLINE");
			break;
		case ICQ_USER_OFFLINE:
			strcpy(szBuffer, "USER_OFFLINE");
			break;
		case ICQ_USER_FOUND:
			strcpy(szBuffer, "USER_FOUND");
			break;
		case ICQ_RECEIVE_MESSAGE:
			strcpy(szBuffer, "RECEIVE_MESSAGE");
			break;
		case ICQ_END_OF_SEARCH:
			strcpy(szBuffer, "END_OF_SEARCH");
			break;
		case ICQ_INFO_REPLY:
			strcpy(szBuffer, "INFO_REPLY");
			break;
		case ICQ_EXT_INFO_REPLY:
			strcpy(szBuffer, "EXT_INFO_REPLY");
			break;
		case ICQ_STATUS_UPDATE:
			strcpy(szBuffer, "STATUS_UPDATE");
			break;
		case ICQ_MULTIPLE_PACKET:
			strcpy(szBuffer, "MULTIPLE_PACKET");
			break;
		case ICQ_INSTANT_MESSAGE:
			strcpy(szBuffer, "INSTANT_MESSAGE");
			break;

		case ICQ_REPLY_X1:
			strcpy(szBuffer, "REPLY_X1");
			break;
		case ICQ_REPLY_X2:
			strcpy(szBuffer, "REPLY_X2");
			break;
		case ICQ_UPDATE_REPLY:
			strcpy(szBuffer, "UPDATE_REPLY");
			break;
		case ICQ_UPDATE_EXT_REPLY:
			strcpy(szBuffer, "UPDATE_EXT_REPLY");
			break;
		case ICQ_NEW_USER_UIN:
			strcpy(szBuffer, "NEW_USER_UIN");
			break;
		case ICQ_NEW_USER_REPLY:
			strcpy(szBuffer, "NEW_USER_REPLY");
			break;
		case ICQ_QUERY_REPLY:
			strcpy(szBuffer, "QUERY_REPLY");
			break;
		case ICQ_SYSTEM_MESSAGE:
			strcpy(szBuffer, "SYSTEM_MESSAGE");
			break;

		default:
			sprintf(szBuffer, "Unknown: (%04lx)", nCommand);
	}
}

// Routines to handle loading/saving message logs to disk...

// Read a set of ICQMsgs from disk
// Returns first message in list (if any);
ICQMsg *
ReadICQMessages(char *szFileName) {
	ICQMsg *pFirstICQMsg = NULL, *pICQMsg;
    HANDLE hFile;

	// Open the file for reading...

	hFile = CreateFile(szFileName,    // open the file 
		GENERIC_READ,                 // open for reading 
		0,                            // do not share 
		NULL,                         // no security 
		OPEN_EXISTING,                // existing file only 
		FILE_ATTRIBUTE_NORMAL,        // normal file 
		NULL);                        // no attr. template 
	if (hFile == INVALID_HANDLE_VALUE) { 
		return NULL;
	} 

	// Read in messages, one at a time...
	while(pICQMsg = ReadICQMsg(hFile)) {
		AddICQMsgTail(&pFirstICQMsg, pICQMsg);
	}

	CloseHandle(hFile); 
    return pFirstICQMsg;

}

// Write a set of ICQMsgs to disk
BOOL
WriteICQMessages(char *szFileName, ICQMsg *pICQMsg) {
    HANDLE hFile;

	// Open the file for reading...

	hFile = CreateFile(szFileName,    // open the file 
		GENERIC_WRITE,                // open for writing 
		0,                            // do not share 
		NULL,                         // no security 
		CREATE_ALWAYS,                // existing file only 
		FILE_ATTRIBUTE_NORMAL,        // normal file 
		NULL);                        // no attr. template 
	if (hFile == INVALID_HANDLE_VALUE) { 
		return FALSE;
	} 

	// Write messages, one at a time...
	while(pICQMsg && WriteICQMsg(hFile, pICQMsg)) {
		pICQMsg = pICQMsg->pNextMsg;
	}

	CloseHandle(hFile); 
    return TRUE;

}

// Read ICQMsg info from disk
ICQMsg *
ReadICQMsg(HANDLE hFile) {
	StaticICQMsg sim;
	DWORD  dwBytesRead;
	ICQMsg *pICQMsg;

	if (ReadFile(hFile, &sim, sizeof(sim), &dwBytesRead, NULL) == FALSE)
		return NULL;

	if (dwBytesRead != sizeof(sim))
		return NULL;

	// Do a quick sanity check on the data...
	if (sim.nMsgLen > 1024) return FALSE;

	// Attempt to allocate ICQMsg object...
	pICQMsg = CreateICQMsg(sim.nMsgLen,
						  (sim.bServerMsg) ? TRUE : FALSE  // Doing this isn't strictly necessary, but it ensures that bool is *exactly* TRUE or FALSE
						  );

	if (pICQMsg == FALSE) return NULL;

	// We have data in sim structure.
	// We need to read in the following fields:
	// pData

	if (ReadFile(hFile, pICQMsg->pData, sim.nMsgLen, &dwBytesRead, NULL) == FALSE) {
		DestroyICQMsg(pICQMsg);
		return NULL;
	}
	if (dwBytesRead != sim.nMsgLen) {
		DestroyICQMsg(pICQMsg);
		return NULL;
	}

	// Now process the message
	ProcessICQMsg(pICQMsg, NULL);

	// Now copy across info that we need to restore (there's probably a better way of doing this..)
	strcpy(pICQMsg->szFrom, sim.szFrom);
	strcpy(pICQMsg->szAnnotation, sim.szAnnotation);
	pICQMsg->lTime = sim.lTime;

	// And return
	return pICQMsg;
}

// Write an ICQMsg to disk
BOOL
WriteICQMsg(HANDLE hFile, ICQMsg *pICQMsg) {
	StaticICQMsg sim;
	DWORD  dwBytesRead;


	// Now copy across info that we need to restore (there's probably a better way of doing this..)
	sim.nMsgLen = pICQMsg->nMsgLen;
	sim.bServerMsg = pICQMsg->bServerMsg;
	strcpy(sim.szFrom, pICQMsg->szFrom);
	strcpy(sim.szAnnotation, pICQMsg->szAnnotation);
	sim.lTime = pICQMsg->lTime;

	if (WriteFile(hFile, &sim, sizeof(sim), &dwBytesRead, NULL) == FALSE)
		return FALSE;

	if (dwBytesRead != sizeof(sim))
		return FALSE;

	// We need to write the following fields:
	// pData
	if (WriteFile(hFile, pICQMsg->pData, sim.nMsgLen, &dwBytesRead, NULL) == FALSE) {
		DestroyICQMsg(pICQMsg);
		return FALSE;
	}
	if (dwBytesRead != sim.nMsgLen) {
		DestroyICQMsg(pICQMsg);
		return FALSE;
	}

	return TRUE;
}