#ifndef __FFXMSG_H__
#define __FFXMSG_H__

//////////////////////////////////////////////////////////////////
//
// ICQMsg.h - ICQ Message processing defines.
// 
//   Copyright (C) 1998 Paul Holden
//   paulh@javigate.com
// 
//   File Created: 6  Nov 1998
//   Last Update: 16 Nov 1998
//
//   For more info and more recent version see:
//     http://www.javigate.com/ICQ/
//
//  Standard Disclaimer:
//  --------------------
//    Note that I am in no way affiliate with Mirabilis. I have found all these
//    information by tracing UDP packets (this complies with Mirabilis License
//    agreement). This information are unofficial and may be incorrect.
//
//    This program is provided "as is" and comes with no warranty,
//    express if implied. Use this program at you own risk. Yadda Yadda.


#include <time.h>

unsigned char icq_check_data4[256];
unsigned char icq_check_data5[256];

// Client packets are like this for v5
//0  WORD  nVersion;
//2  DWORD nZero;
//6  DWORD nUIN;
//10 DWORD nSession;
//14 WORD  nCommand;
//16 WORD  nSeqNum1;
//18 WORD  nSeqNum2;
//20 DWORD nCheckCode;

// Client packets are like this for v4
//0 WORD   04 00                 VERSION
//2 DWORD  xx xx 00 00           RANDOMNUM
//6 WORD   xx xx                 COMMAND
//8 WORD   xx xx                 SEQ_NUM1
//10 WORD  xx xx                 SEQ_NUM2
//12 DWORD xx xx xx xx           UIN
//20 DWORD xx xx xx xx           CHECKCODE


// Server packets are like this for v5
//0  WORD  nVersion
//2  BYTE  nZero
//3  DWORD nSession
//7  WORD  nCommand
//9  WORD  nSeqNum1
//11 WORD  nSeqNum2
//13 DWORD nUIN
//17 DWORD nCheckCode

// Our structure used to hold message info
// Used in a linked list
typedef struct ICQMsg_struct {

	struct ICQMsg_struct *pNextMsg, *pPrevMsg;

	unsigned long nMsgLen;
	unsigned char *pData;
	unsigned char *pDecryptedData;
	char		  szFrom[32];
	char		  szExtraInfo[256];
	char		  szAnnotation[256];
	BOOL		  bServerMsg;
	BOOL		  bDecrypted;
	time_t		  lTime;

} ICQMsg;

// This structure is used to save message info to disk.
// It needs to be flat (i.e. with no pointers..) so I've
// called it static. Probably not the right terminology, but
// who cares?
// I've used DWORDs, WORDS and BYTEs as these are 
// easy to use on other platforms (4/2/1 bytes)
// This structure contains all the info that can't be recreated
// e.g. from, time, annotation etc
typedef struct {
	DWORD	nMsgLen;
	BYTE    bServerMsg;
	char    szFrom[32];
	char    szAnnotation[256];
	time_t  lTime;

} StaticICQMsg;


//////////////////////////////////////////////////////////////
// * = Client -> Server
// % = Server -> Client
#define ICQ_ACK				0x000A		//*%Acknowledgement
#define ICQ_SEND_MESSAGE	0x010E		//*Send message through server (to offline user)
#define ICQ_LOGIN			0x03E8		//*Login on server
#define ICQ_CONTACT_LIST	0x0406		//*Inform the server of my contact list
#define ICQ_SEARCH_UIN		0x041A		//*Search for user using his/her UIN
#define ICQ_SEARCH_USER		0x0424		//*Search for user using his/her name or e-mail
#define ICQ_KEEP_ALIVE		0x042E		//*Sent to indicate connection is still up
#define ICQ_SEND_TEXT_CODE	0x0438		//*Send special message to server as text
#define ICQ_LOGIN_1			0x044C		//*Sent during login
#define ICQ_INFO_REQ		0x0460		//*Request basic information about a user
#define ICQ_EXT_INFO_REQ	0x046A		//*Request extended information about a user
#define ICQ_CHANGE_PASSWORD 0x049C		//*Change the user's password
#define ICQ_STATUS_CHANGE	0x04D8		//*User has changed online status (Away etc)
#define ICQ_LOGIN_2			0x0528		//*Sent during login

#define ICQ_UPDATE_INFO		0x050A		//*Update my basic information
#define ICQ_UPDATE_EXT_INFO 0x04B0		//*Update my extended information
#define ICQ_ADD_TO_LIST		0x053C      //*Add user to my contact list
#define ICQ_REQ_ADD_TO_LIST 0x0456		//*Request authorization to add to contact list
#define ICQ_QUERY_SERVERS	0x04BA		//*Query the server about address to other servers
#define ICQ_QUERY_ADDONS	0x04C4      //*Query the server about globally defined add-ons
#define ICQ_NEW_USER_1		0x04EC      //*Ask for permission to add a new user
#define ICQ_NEW_USER_REG	0x03FC      //*Register a new user
#define ICQ_NEW_USER_INFO	0x04A6		//*Send basic information about a new user
#define ICQ_CMD_X1			0x0442      //*Unknown
// This has the same 
//#define ICQ_MSG_TO_NEW_USER 0x0456	//*Send a message to a user not on my contact list



// Server stuff
#define ICQ_LOGIN_REPLY		0x005A		//%Login reply
#define ICQ_USER_ONLINE		0x006E		//%User on contact list is online/has changed online status
#define ICQ_USER_OFFLINE	0x0078		//%User on contact list has gone offline
#define ICQ_USER_FOUND		0x008C		//%User record found matching search criteria
#define ICQ_RECEIVE_MESSAGE 0x00DC		//%Message sent while offline/through server
#define ICQ_END_OF_SEARCH	0x00A0		//%No more USER_FOUND will be sent
#define ICQ_INFO_REPLY		0x0118		//%Return basic information about a user
#define ICQ_EXT_INFO_REPLY	0x0122		//%Return extended information about a user
#define ICQ_STATUS_UPDATE	0x01A4		//%User on contact list has changed online status (Away etc)
#define ICQ_MULTIPLE_PACKET 0x0212		//%Multiple packets folded into one...
#define ICQ_INSTANT_MESSAGE 0x0104		//%Some kind of message from user via server...

#define ICQ_REPLY_X1		0x021C		//%Unknown (returned during login)
#define ICQ_REPLY_X2		0x00E6		//%Unknown (confirm my UIN?)
#define ICQ_UPDATE_REPLY	0x01E0		//%Confirmation of basic information update
#define ICQ_UPDATE_EXT_REPLY 0x00C8		//%Confirmation of extended information update
#define ICQ_NEW_USER_UIN	0x0046		//%Confirmation of creation of new user and newly assigned UIN
#define ICQ_NEW_USER_REPLY	0x00B4		//%Confirmation of new user basic information
#define ICQ_QUERY_REPLY		0x0082		//%Response to QUERY_SEVERS or QUERY_ADDONS
#define ICQ_SYSTEM_MESSAGE	0x01C2		//%System message with URL'ed button


// Function prototypes. Not all of them... :-)
ICQMsg *CreateICQMsg(unsigned long, BOOL);
void    DestroyICQMsg(ICQMsg *pICQMsg);
void    DestroyICQMsgList(ICQMsg **);
void    AddICQMsgHead(ICQMsg **, ICQMsg *);
void    AddICQMsgTail(ICQMsg **, ICQMsg *);
void    RemoveICQMsg (ICQMsg **, ICQMsg *);

ICQMsg *FindICQMsg(ICQMsg **, DWORD, WORD, WORD);

void DecryptPacket(ICQMsg *);
void GetCommandName(WORD, char *);
void GetICQMsgTime(ICQMsg *, char *);
void GetICQMsgInfo(ICQMsg *, char *);
void GetICQMsgHeaderInfo(ICQMsg **, ICQMsg *, char *);
DWORD GetKey(ICQMsg *);
DWORD Reverse(DWORD);

BOOL ProcessICQMsg(ICQMsg *, struct sockaddr_in *);
void ProcessLogin(ICQMsg *);
void ProcessAck(ICQMsg *);
void ProcessContactList(ICQMsg *);
void ProcessInstantMessage(ICQMsg *);
void ProcessTextCode(ICQMsg *);

void RespondToMessage(ICQMsg *, struct sockaddr_in *, char *);

WORD  ICQVersion(ICQMsg *);
DWORD ICQZero(ICQMsg *);
DWORD ICQUIN(ICQMsg *);
DWORD ICQSession(ICQMsg *);
WORD  ICQCommand(ICQMsg *);
WORD  ICQSeqNum1(ICQMsg *);
WORD  ICQSeqNum2(ICQMsg *);
DWORD ICQCheckCode(ICQMsg *);

ICQMsg *ReadICQMessages(char *);
BOOL    WriteICQMessages(char *, ICQMsg *);
ICQMsg *ReadICQMsg(HANDLE);
BOOL    WriteICQMsg(HANDLE, ICQMsg *);

#endif